#include "mpc.h"
#include <stdio.h>
#include <stdlib.h>
#include "test.h"

char bottom;

#define MAX_REQ 16
#define DEF_MAX_MSG 2000000
/* 
   This program tests a flood of data for both unexpected and expected messages   
   to test any internal message fragmentation or protocol shifts

   An optional argument can change the maximum message size.  For example, use
      flood 9000000
   to stress the memory system (the size is the number of ints, not bytes)
 */

void SetupData (int *, int, int);
void SetupRdata (int *, int);
int CheckData (int *, int, int, MPC_Status *);

#ifdef VERBOSE
static int verbose = 1;
#else
static int verbose = 0;
#endif

int
main (int argc, char **argv)
{
  MPC_Comm comm;
  MPC_Request r[MAX_REQ];
  MPC_Status s[MAX_REQ];
  int msgsize, maxmsg, root, i, size, rank, err = 0, toterr;
  int max_msg_size = DEF_MAX_MSG;
  int *sbuf, *rbuf;

  MPC_Init (&argc, &argv);

  comm = MPC_COMM_WORLD;

  MPC_Comm_size (comm, &size);
  MPC_Comm_rank (comm, &rank);

  if (size < 2)
    {
      printf ("This test requires at least 2 processors\n");
      MPC_Abort (comm, 1);
    }

  /* Check for a max message argument */
  if (rank == 0)
    {
      if (argc > 1)
	{
	  max_msg_size = atoi (argv[1]);
	  /* Correct if unrecognized argument */
	  if (max_msg_size <= 0)
	    max_msg_size = DEF_MAX_MSG;
	}
    }
  MPC_Bcast (&max_msg_size, 1, MPC_INT, 0, MPC_COMM_WORLD);

  /* First, try large blocking sends to root */
  root = 0;

  msgsize = 128;
  maxmsg = max_msg_size;
  if (rank == root)
    printf ("Blocking sends: ");
  while (msgsize < maxmsg)
    {
      if (rank == root)
	{
	  if (verbose)
	    {
	      printf ("%d ", msgsize);
	      fflush (stdout);
	    }
	  rbuf = (int *) malloc (msgsize * sizeof (int));
	  if (!rbuf)
	    {
	      printf ("Could not allocate %d words\n", msgsize);
	      MPC_Abort (comm, 1);
	    }
	  for (i = 0; i < size; i++)
	    {
	      if (i == rank)
		continue;
	      SetupRdata (rbuf, msgsize);
	      MPC_Recv (rbuf, msgsize, MPC_INT, i, 2 * i, comm, s);
	      if (verbose)
		{
		  printf ("check %d ", i);
		  fflush (stdout);
		}
	      err += CheckData (rbuf, msgsize, 2 * i, s);
	    }
	  free (rbuf);
	}
      else
	{
	  sbuf = (int *) malloc (msgsize * sizeof (int));
	  if (!sbuf)
	    {
	      printf ("Could not allocate %d words\n", msgsize);
	      MPC_Abort (comm, 1);
	    }
	  SetupData (sbuf, msgsize, 2 * rank);
	  MPC_Send (sbuf, msgsize, MPC_INT, root, 2 * rank, comm);
	  free (sbuf);
	}
      msgsize *= 4;
    }
  if (rank == 0)
    {
      printf ("\n");
      fflush (stdout);
    }

  /* Next, try unexpected messages with Isends */
  msgsize = 128;
  maxmsg = max_msg_size;
  if (rank == root)
    printf ("Unexpected recvs: ");
  while (msgsize < maxmsg)
    {
      if (rank == root)
	{
	  if (verbose)
	    {
	      printf ("%d ", msgsize);
	      fflush (stdout);
	    }
	  rbuf = (int *) malloc (msgsize * sizeof (int));
	  if (!rbuf)
	    {
	      printf ("Could not allocate %d words\n", msgsize);
	      MPC_Abort (comm, 1);
	    }
	  MPC_Barrier (comm);
	  for (i = 0; i < size; i++)
	    {
	      if (i == rank)
		continue;
	      SetupRdata (rbuf, msgsize);
	      MPC_Recv (rbuf, msgsize, MPC_INT, i, 2 * i, comm, s);
	      err += CheckData (rbuf, msgsize, 2 * i, s);
	    }
	  free (rbuf);
	}
      else
	{
	  sbuf = (int *) malloc (msgsize * sizeof (int));
	  if (!sbuf)
	    {
	      printf ("Could not allocate %d words\n", msgsize);
	      MPC_Abort (comm, 1);
	    }
	  SetupData (sbuf, msgsize, 2 * rank);
	  MPC_Isend (sbuf, msgsize, MPC_INT, root, 2 * rank, comm, r);
	  MPC_Barrier (comm);
	  MPC_Wait (r, s);
	  free (sbuf);
	}
      msgsize *= 4;
    }
  if (rank == 0)
    {
      printf ("\n");
      fflush (stdout);
    }

  /* Try large synchronous blocking sends to root */
  root = 0;

  msgsize = 128;
  maxmsg = max_msg_size;
  if (rank == root)
    printf ("Synchronous sends: ");
  while (msgsize < maxmsg)
    {
      if (rank == root)
	{
	  if (verbose)
	    {
	      printf ("%d ", msgsize);
	      fflush (stdout);
	    }
	  rbuf = (int *) malloc (msgsize * sizeof (int));
	  if (!rbuf)
	    {
	      printf ("Could not allocate %d words\n", msgsize);
	      MPC_Abort (comm, 1);
	    }
	  for (i = 0; i < size; i++)
	    {
	      if (i == rank)
		continue;
	      SetupRdata (rbuf, msgsize);
	      MPC_Recv (rbuf, msgsize, MPC_INT, i, 2 * i, comm, s);
	      err += CheckData (rbuf, msgsize, 2 * i, s);
	    }
	  free (rbuf);
	}
      else
	{
	  sbuf = (int *) malloc (msgsize * sizeof (int));
	  if (!sbuf)
	    {
	      printf ("Could not allocate %d words\n", msgsize);
	      MPC_Abort (comm, 1);
	    }
	  SetupData (sbuf, msgsize, 2 * rank);
	  MPC_Ssend (sbuf, msgsize, MPC_INT, root, 2 * rank, comm);
	  free (sbuf);
	}
      msgsize *= 4;
    }
  if (rank == 0)
    {
      printf ("\n");
      fflush (stdout);
    }

  /* Next, try expected messages with Rsend */
  msgsize = 128;
  maxmsg = max_msg_size;
  if (rank == root)
    printf ("Expected recvs and Rsend: ");
  while (msgsize < maxmsg)
    {
      if (rank == root)
	{
	  if (verbose)
	    {
	      printf ("%d ", msgsize);
	      fflush (stdout);
	    }
	  rbuf = (int *) malloc (msgsize * sizeof (int));
	  if (!rbuf)
	    {
	      printf ("Could not allocate %d words\n", msgsize);
	      MPC_Abort (comm, 1);
	    }
	  for (i = 0; i < size; i++)
	    {
	      if (i == rank)
		continue;
	      SetupRdata (rbuf, msgsize);
	      MPC_Irecv (rbuf, msgsize, MPC_INT, i, 2 * i, comm, r);
	      MPC_Send (MPC_BOTTOM, 0, MPC_INT, i, 2 * i + 1, comm);
	      MPC_Wait (r, s);
	      err += CheckData (rbuf, msgsize, 2 * i, s);
	    }
	  free (rbuf);
	}
      else
	{
	  sbuf = (int *) malloc (msgsize * sizeof (int));
	  if (!sbuf)
	    {
	      printf ("Could not allocate %d words\n", msgsize);
	      MPC_Abort (comm, 1);
	    }
	  SetupData (sbuf, msgsize, 2 * rank);
	  MPC_Recv (MPC_BOTTOM, 0, MPC_INT, root, 2 * rank + 1, comm, s);
	  MPC_Rsend (sbuf, msgsize, MPC_INT, root, 2 * rank, comm);
	  free (sbuf);
	}
      msgsize *= 4;
    }
  if (rank == 0)
    {
      printf ("\n");
      fflush (stdout);
    }


  MPC_Allreduce (&err, &toterr, 1, MPC_INT, MPC_SUM, MPC_COMM_WORLD);

  if (rank == 0)
    {
      if (toterr == 0)
	printf (" No Errors\n");
      else
	printf ("!! found %d errors\n", toterr);
    }
  if (toterr)
    {
      printf ("!! found %d errors on processor %d\n", err, rank);
    }

  MPC_Finalize ();
  return toterr;
}

void
SetupData (sbuf, n, tag)
     int *sbuf, n, tag;
{
  int i;

  for (i = 0; i < n; i++)
    sbuf[i] = i;
}

int
CheckData (rbuf, n, tag, s)
     int *rbuf, n, tag;
     MPC_Status *s;
{
  int act_n, i;

  MPC_Get_count (s, MPC_INT, &act_n);
  if (act_n != n)
    {
      printf ("Received %d instead of %d ints\n", act_n, n);
      return 1;
    }
  for (i = 0; i < n; i++)
    {
      if (rbuf[i] != i)
	{
	  printf ("rbuf[%d] is %d, should be %d\n", i, rbuf[i], i);
	  return 1;
	}
    }
  return 0;
}

void
SetupRdata (rbuf, n)
     int *rbuf, n;
{
  int i;

  for (i = 0; i < n; i++)
    rbuf[i] = -(i + 1);
}
