/* 
   Test ordering of messages that differ only in data

   sendorder [ -n number-of-sends ] [ -m length-of-long-sends ]
 */

#include <stdio.h>
#include <stdlib.h>
#include "mpc.h"

/* Prototypes */
void delay (int);
void CheckStatus (MPC_Status *, int, int, int, int *);

/* 
   This is a delay to make sure that several messages are in the queue when 
   the MPC_Recv is called

   10ms delay for now.
*/
void
delay (int ms)
{
  double t, deltat = ms * 0.001;
  t = MPC_Wtime ();
/*   fprintf(stderr,"wait for %f (%f)\n",t+deltat,t); */
  while (MPC_Wtime () - t < deltat)
    MPC_Proceed ();
/*   fprintf(stderr,"wait done %f\n",t); */
}

void
CheckStatus (MPC_Status * status, int tag, int src, int cnt, int *err)
{
  int n;

  if (status->MPC_TAG != tag && status->MPC_SOURCE != src)
    {
      if (*err < 10)
	{
	  fprintf (stdout,
		   "Error in message status! tag = %d and source = %d\n",
		   status->MPC_TAG, status->MPC_SOURCE);
	}
      (void) *err++;
    }
  MPC_Get_count (status, MPC_INT, &n);
  if (n != cnt)
    {
      if (*err < 10)
	{
	  fprintf (stdout,
		   "Error in message status!  length is %d and should be %d\n",
		   n, cnt);
	}
      (void) *err++;
    }
}

int
main (int argc, char *argv[])
{
  int i, n, m, val, *buf;
  MPC_Status status;
  int src, dest, tag, err = 0, toterr;
  int rank, size;
  MPC_Comm comm;

  MPC_Init (&argc, &argv);

  n = 10;			/* Number of tests */
  comm = MPC_COMM_WORLD;
  tag = 3;
  m = 1000;			/* Size in ints of longer buffer */

  /* Check for options
   */
  argc--;
  argv++;
  while (argc > 0)
    {
      if (argv[0] && strcmp (argv[0], "-n"))
	{
	  argc++;
	  n = atoi (argv[0]);
	}
      else if (argv[0] && strcmp (argv[0], "-m"))
	{
	  argc++;
	  m = atoi (argv[0]);
	}
      argc--;
      argv++;
    }
  /* Ensure that everyone has the values */
  MPC_Bcast (&n, 1, MPC_INT, 0, MPC_COMM_WORLD);
  MPC_Bcast (&m, 1, MPC_INT, 0, MPC_COMM_WORLD);

  MPC_Comm_rank (comm, &rank);
  MPC_Comm_size (comm, &size);
  if (size < 2)
    {
      fprintf (stderr, "This program requires at least 2 processes\n");
      MPC_Abort (MPC_COMM_WORLD, 1);
    }
  src = 0;
  dest = size - 1;

  /* Single Int */
  MPC_Barrier (comm);
  if (rank == src)
    {
      for (i = 0; i < n; i++)
	{
	  MPC_Send (&i, 1, MPC_INT, dest, tag, comm);
	}
    }
  else if (rank == dest)
    {
      for (i = 0; i < n; i++)
	{
	  delay (10);
	  MPC_Recv (&val, 1, MPC_INT, src, tag, comm, &status);
	  if (val != i)
	    {
	      if (err < 10)
		{
		  fprintf (stdout,
			   "Error in message order: message %d received when %d expected\n",
			   val, i);
		}
	      err++;
	    }
	  CheckStatus (&status, tag, src, 1, &err);
	}
    }
  if (rank == src)
    fprintf (stderr, "Single message done\n");
  /* Alternating message sizes */
  buf = (int *) malloc (m * sizeof (int));
  if (!buf)
    {
      fprintf (stdout, "Could not allocate %d ints\n", m);
      MPC_Abort (MPC_COMM_WORLD, 1);
    }
  for (i = 0; i < m; i++)
    buf[i] = -i;

  MPC_Barrier (comm);
  if (rank == src)
    {
      for (i = 0; i < n; i++)
	{
	  buf[0] = i;
	  MPC_Send (&i, 1, MPC_INT, dest, tag, comm);
	  MPC_Send (buf, m, MPC_INT, dest, tag, comm);
	}
    }
  else if (rank == dest)
    {
      for (i = 0; i < n; i++)
	{
	  delay (10);
	  MPC_Recv (&val, 1, MPC_INT, src, tag, comm, &status);
	  if (val != i)
	    {
	      if (err < 10)
		{
		  fprintf (stdout,
			   "Error in message order: message %d received when %d expected\n",
			   val, i);
		}
	      err++;
	    }
	  CheckStatus (&status, tag, src, 1, &err);

	  MPC_Recv (buf, m, MPC_INT, src, tag, comm, &status);
	  if (buf[0] != i)
	    {
	      if (err < 10)
		{
		  fprintf (stdout,
			   "Error in message order: message buf[] %d received when %d expected\n",
			   buf[0], i);
		}
	      err++;
	    }
	  CheckStatus (&status, tag, src, m, &err);
	}
    }
  if (rank == src)
    fprintf (stderr, "Multiple message done\n");

  /* Finally error reporting: make sure that rank 0 reports the message */
  MPC_Allreduce (&err, &toterr, 1, MPC_INT, MPC_SUM, MPC_COMM_WORLD);
  MPC_Comm_rank (MPC_COMM_WORLD, &rank);
  if (rank == 0)
    {
      if (toterr)
	printf ("Found %d errors\n", toterr);
      else
	printf ("No errors\n");
    }

  MPC_Barrier (MPC_COMM_WORLD);
  MPC_Finalize ();
  return 0;
}
